# *-* coding=utf8 *-*
# ******************************************************************************
# wirkungsgradkennfeld.py
#
# Alles was zum Erstellen des Wirkungsgradkennfeld notwendig ist, befindet
# sich in diesem Modul.
# ------------------------------------------------------------------------------
#
# Copyright (c) 2018 JSH <downloader@gmx.de>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
# ******************************************************************************

import sqlite3
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt

from functools import reduce
from collections import Counter


__all__ = ['Wirkungsgrad']


class Wirkungsgrad(object):
    """Erstellt ein Wirkungsgradkennfeld aus Hybrid Assistant Daten"""

    def __init__(self, database='hybridassistant.db'):
        """Initialisiert das Objekt

        :param database: Pfad zur Datenbank
        """
        self.database = database
        self.last_min = 1
        self.last_max = 100
        self.rpm_min = 1
        self.rpm_max = 6000
        self.wirkungsgrad_min = 1
        self.wirkungsgrad_max = 40
        self.notwendige_vorkommen = 5
        self.heizwert_des_kraftstoffs = 11.4
        self.titel = "Wirkungsgrad"

    def erzeuge_diagramm(self, pfad='Wirkungsgrad.png'):
        """Erstellt Wirkungsgradkennfeld

        :param pfad: Pfad der Ausgabedatei (.[eps,pdf,pgf,png,ps,raw,rgba,svg,svgz])
        :return: Im Fehlerfall eine Message, sonst ""
        """
        self._validate_parameter()
        (dbdaten, resultmsg) = ermittel_load_rpm_bsfc(self.database)
        if dbdaten:
            wirkungsgrade = erstelle_hashtable_fuer_last_umdrehung_wirkungsgrade(
                    dbdaten,
                    self.heizwert_des_kraftstoffs
            )
            wirkungsgrade_gemittelt = ermittel_mittelwert(
                wirkungsgrade,
                self.notwendige_vorkommen
            )
            matrix = erstelle_matrix_der_daten(
                data=wirkungsgrade_gemittelt,
                last_min=self.last_min,
                last_max=self.last_max,
                rpm_min=self.rpm_min,
                rpm_max=self.rpm_max
            )
            resultmsg = self._erzeuge_contoutf(matrix, pfad)
        return resultmsg

    # ---------
    #  Private
    # ---------
    def _validate_parameter(self):
        """Prüft ob alle Attribute im gültigen Bereich liegen
        """
        msg = "Fehlerhafter Parameter: "
        assert 0 < self.last_min <= 100, msg+"last_min=[1-100]"
        assert 0 < self.last_max <= 100, msg+"last_max=[1-100]"
        assert 0 < self.rpm_min <= 7000, msg+"rmp_min=[1-7000]"
        assert 0 < self.rpm_max <= 7000, msg+"rmp_max=[1-7000]"
        assert 0 < self.wirkungsgrad_min <= 100, msg+"wirkungsgrad_min=[1-100]"
        assert 0 < self.wirkungsgrad_max <= 100, msg+"wirkungsgrad_max=[1-100]"
        assert 0 < self.notwendige_vorkommen <= 1000, msg+"notwendige_vorkommen=[1-1000]"
        assert 0 < self.heizwert_des_kraftstoffs <= 100, msg+"heizwert_des_kraftstoffs=[1.0-100.0]"

    def _erzeuge_contoutf(self, matrix, pfad):
        """Erzeugt Diagramm und gibt es in einer Datei aus

        :param matrix: numpy-array[N,M] mit allen Wirkungsgraden
        :param pfad: Ausgabepfad der Datei
        :return: Im Fehlerfall eine Message, sonst ""
        """
        resultmsg = ""
        x, y = np.meshgrid(
            np.array(range(self.rpm_min, self.rpm_max + 10, 10)),
            np.array(range(self.last_max, self.last_min, -1))
        )
        diagramm = plt.contourf(x, y, matrix,
                                cmap=plt.cm.get_cmap('winter'),
                                vmin=self.wirkungsgrad_min,
                                levels=[x for x in range(0, self.wirkungsgrad_max + 1, 2)]
                                )
        plt.title(self.titel)
        plt.xlabel('Umdrehungen pro Minute')
        plt.ylabel('Last in Prozent')
        plt.colorbar(diagramm, extend='both', label='Wirkungsgrad in Prozent')
        try:
            plt.savefig(pfad)
        except ValueError as e:
            resultmsg = e
        return resultmsg


# ---------
#  HELPERS
# ---------
def ermittel_load_rpm_bsfc(database):
    """Liefert ICE_LOAD, ICE_RPM und BSFC

    :param database: Pfad zur Datenbank
    :return: (sqlite3.Cursor, errormessage)
    """
    resultmsg = ""
    result = None
    conn = sqlite3.connect(database)
    c = conn.cursor()
    try:
        result = c.execute("""
            select ICE_LOAD,ICE_RPM,BSFC 
                from FASTLOG 
                WHERE ICE_RPM > 800 and BSFC>0 and ICE_LOAD>0
                ORDER BY ICE_LOAD, ICE_RPM
            """
                           )
    except sqlite3.OperationalError as e:
        resultmsg = "Datenbank entspricht nicht den Anforderungen: '{}'".format(e)
    return result, resultmsg


def erstelle_hashtable_fuer_last_umdrehung_wirkungsgrade(dbcursor, heizwert_des_kraftstoffs):
    """Liefert Hashtable im Format (load,rpm)=[wirkungsgrade,...]

    :param dbcursor: DBCursor für ICE_LOAD, ICE_RPM und BSFC
    :param heizwert_des_kraftstoffs: in kWh/kg
    :return: Hashtable (load,rpm)=[wirkungsgrade,...]
    """
    data = {}
    for load, rpm, bsfc in dbcursor:
        key = (load, rpm)
        if key in data:
            data[key].append(berechne_wirkungsgrad(bsfc, heizwert_des_kraftstoffs))
        else:
            data[key] = [berechne_wirkungsgrad(bsfc, heizwert_des_kraftstoffs)]
    return data


def ermittel_mittelwert(data, mindeszahl_der_vorkommen):
    """Liefert den Mittelwert der übergebenen Liste

    :param data: Liste mit Werten
    :param mindeszahl_der_vorkommen: Ab wann soll gezählt werden?
    :return: Hashtable (load,rpm)=wirkungsgrad
    """
    newdata = {}
    for key in data:
        values = entferne_zufallsdaten(data[key], mindeszahl_der_vorkommen)
        if values:
            mittelwert = reduce(lambda x, y: x + y, values) / len(values)
            newdata[key] = round(mittelwert)
    return newdata


def entferne_zufallsdaten(data_values, mindeszahl_der_vorkommen):
    """Entfernt alle Werte aus der Liste die weniger als x-mal vorkommen

    Die zurückgelieferten Werte sind nach Anzahl ihrer Vorkommen
    sortiert.

    :param data_values: Liste mit Werten
    :param mindeszahl_der_vorkommen: Ab wann soll gezählt werden?
    :return: Liste mit Werten
    """
    values = []
    c = Counter(data_values)
    for value, counter in c.most_common():
        if counter >= mindeszahl_der_vorkommen:
            for i in range(counter):
                values.append(value)
    return values


def berechne_wirkungsgrad(bsfc, heizwert_des_kraftstoffs):
    """Berechnet anhand des BSFC Werts den Wirkungsgrad

    :param bsfc: BSFC Wert
    :param heizwert_des_kraftstoffs: in kWh/kg
    :return: Wirkungsgrad
    """
    wirkungsgrad = 100 / ((bsfc / 1000) * heizwert_des_kraftstoffs)
    return round(wirkungsgrad)


def erstelle_matrix_der_daten(data, last_min, last_max, rpm_min, rpm_max):
    """Erstellt Datenmatrix

    :param data: Hashtable (load,rpm)=wirkungsgrad
    :param last_min: minimaler Lastindex
    :param last_max: maximaler Lastindes
    :param rpm_min: minimale Umdrehung
    :param rpm_max: maximale Umdrehung
    :return numpy.array[n,m] mit allen Wirkungsgraden
    """
    zeilen = []
    for row in range(last_max, last_min, -1):
        spalten = [0]
        for col in range(rpm_min - (rpm_min % 10), rpm_max, 10):
            key = (row, col)
            if key in data:
                spalten.append(data[key])
            else:
                spalten.append(0)
        zeilen.append(interpoliere_fehlende_daten(spalten))
    return np.array(zeilen)


def interpoliere_fehlende_daten(spalten):
    """Füllt Datenlücken innerhalb der Messbereiche sinnvoll auf

    :param spalten: Liste aller Wirkungsgrade zu einem Lastbereich
    :return pandas.Series aller Wikrungsgrade
    """
    leftpos, rightpos = position_des_wertebereichs_ungleich_null(spalten)
    for i in range(leftpos, rightpos):
        if spalten[i] == 0:
            spalten[i] = np.nan
    return pd.Series(spalten).interpolate()


def position_des_wertebereichs_ungleich_null(liste):
    """Ermittelt die Position des Wertebereichs ungleich 0

    :param liste: Liste mit Werten
    :return: Tuple der Positionen (linker_index, rechter_index)
    """
    links = rechts = 0
    for links in range(0, len(liste)):
        if liste[links] > 0:
            break
    for rechts in range(len(liste)-1, 0, -1):
        if liste[rechts] > 0:
            break
    return links, rechts + 1
