#!/usr/bin/env python3
# *-* coding=utf8 *-*
# ******************************************************************************
# wirkungsgradkennfeld.py
#
# Mit Hilfe der aufgezeichneten Daten des Hybrid Assistant erstellt dieses
# Script ein Wirkungsgradkennfeld.
# ------------------------------------------------------------------------------
#
# Copyright (c) 2018 JSH <downloader@gmx.de>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
# ******************************************************************************

import sys
import argparse

from pathlib import Path
from halib import Wirkungsgrad

VERSION = "v1.0.0"


def parse_arguments():
    """Wertet die übergebenen Parameter aus und liefert sie zurück.

    :return: argparse.Namespace
    """

    parser = argparse.ArgumentParser(
        description="Erstellt ein Wirkungsgradkennfeld über Last und Umdrehung"
    )

    parser.add_argument('-d',
                        action="store",
                        dest="DATABASE",
                        type=str,
                        default="hybridassistant.db",
                        help="Hybrid Assistant Datenbank (Default=hybridassistant.db)"
                        )

    parser.add_argument('-o',
                        action="store",
                        dest="OUTPUT",
                        type=str,
                        default="Wirkungsgrad.png",
                        help="Ausgabepfad für das Diagramm (Default=Wirkungsgrad.png)"
                        )

    parser.add_argument('-l',
                        action="store",
                        dest="LAST_MIN",
                        type=check_range_1_100,
                        default="1",
                        help="Kleinste anzuzeigende Last in %% (Default=1)"
                        )

    parser.add_argument('-L',
                        action="store",
                        dest="LAST_MAX",
                        type=check_range_1_100,
                        default="100",
                        help="Größte anzuzeigende Last in %% (Default=100)"
                        )

    parser.add_argument('-r',
                        action="store",
                        dest="RPM_MIN",
                        type=check_range_1_7000,
                        default="900",
                        help="Kleinste anzuzeigende Umdrehung/min (Default=900)"
                        )

    parser.add_argument('-R',
                        action="store",
                        dest="RPM_MAX",
                        type=check_range_1_7000,
                        default="6000",
                        help="Größte anzuzeigende Umdrehung/min (Default=6000)"
                        )

    parser.add_argument('-w',
                        action="store",
                        dest="WIRK_MIN",
                        type=check_range_1_100,
                        default="1",
                        help="Kleinster anzuzeigender Wirkungsgrad in %% (Default=1)"
                        )

    parser.add_argument('-W',
                        action="store",
                        dest="WIRK_MAX",
                        type=check_range_1_100,
                        default="100",
                        help="Größter anzuzeigender Wirkungsgrad in %% (Default=100)"
                        )

    parser.add_argument('-n',
                        action="store",
                        dest="ANZAHL_TREFFER",
                        type=check_range_1_10000,
                        default="5",
                        help="Übereinstimmungen für glaubwürdigen Wert (Default=5)"
                        )

    parser.add_argument('-H',
                        action="store",
                        dest="HEIZWERT",
                        type=check_float_range_1_100,
                        default="11.4",
                        help="Heizwert des Kraftstoffs in kWh/kg (Default=11.4)"
                        )

    parser.add_argument('-T',
                        action="store",
                        dest="TITEL",
                        type=str,
                        default="Wirkungsgrad",
                        help="Überschrift des Diagramms (Default=Wirkungsgrad)"
                        )

    return parser.parse_args()


def check_range_1_100(value):
    """Prüft ob Wert zwischen 1 und 100 liegt

    :param value: Wert
    :return: True/False
    """
    return check_int_range(value, 1, 100)


def check_range_1_7000(value):
    """Prüft ob Wert zwischen 1 und 7000 liegt

    :param value: Wert
    :return: True/False
    """
    return check_int_range(value, 1, 7000)


def check_range_1_10000(value):
    """Prüft ob Wert zwischen 1 und 10000 liegt

    :param value: Wert
    :return: True/False
    """
    return check_int_range(value, 1, 10000)


def check_float_range_1_100(value):
    """Prüft ob Float Wert zwischen 1.0 und 100.0 liegt

    :param value: Wert
    :return: True/False
    """
    value = float(value)
    if value < 1 or value > 100:
        msg = "=>{}<= Es sind nur Werte zwischen 1.0 und 100.0 erlaubt!".format(value)
        raise argparse.ArgumentTypeError(msg)
    return value


def check_int_range(value, von, bis):
    """Prüft ob Wert zwischen 'von' und 'bis' liegt

    :param value: Wert
    :param von: Beginn des Bereichs
    :param bis: Ende des Bereichs
    :return: True/False
    """
    value = int(value)
    if value < von or value > bis:
        msg = "=>{}<= Es sind nur Werte zwischen {} und {} erlaubt!".format(value, von, bis)
        raise argparse.ArgumentTypeError(msg)
    return value


def validate_arguments(arg):
    """Zusätzliche Prüfungen ob Argumente Sinn ergeben

    Im Fehlerfall wird eine Meldung ausgegeben und die Verarbeitung
    wird mit einem ReturnCode > 0 beendet.

    :param arg: argparse.Namespace
    """
    if not Path(arg.DATABASE).is_file():
        stop(1, "Die Datenbank {} existiert nicht!".format(arg.DATABASE))
    if arg.LAST_MIN >= arg.LAST_MAX:
        stop(2, "Der Lastbereich [{}-{}] ergibt keinen Sinn!".format(arg.LAST_MIN, arg.LAST_MAX))
    if arg.RPM_MIN >= arg.RPM_MAX:
        stop(3, "Der Drehzahlbereich [{}-{}] ergibt keinen Sinn!".format(arg.RPM_MIN, arg.RPM_MAX))
    if arg.WIRK_MIN >= arg.WIRK_MAX:
        stop(4, "Der Wirkungsgradbereich [{}-{}] ergibt keinen Sinn!".format(arg.WIRK_MIN, arg.WIRK_MAX))


def stop(exitcode, msg):
    """Beendet mit ReturnCode > 0 und Fehlermeldung die Verarbeitung

    :param exitcode: ReturnCode
    :param msg: Fehlermeldung
    """
    if exitcode > 0:
        msg = "FEHLER: {}".format(msg)
    print(msg, file=sys.stderr)
    sys.exit(exitcode)


def main():
    """Hier findet die Hauptverarbeitung statt
    """

    argument = parse_arguments()
    validate_arguments(argument)

    w = Wirkungsgrad(database=argument.DATABASE)
    w.last_min = argument.LAST_MIN
    w.last_max = argument.LAST_MAX
    w.rpm_min = argument.RPM_MIN
    w.rpm_max = argument.RPM_MAX
    w.wirkungsgrad_min = argument.WIRK_MIN
    w.wirkungsgrad_max = argument.WIRK_MAX
    w.notwendige_vorkommen = argument.ANZAHL_TREFFER
    w.heizwert_des_kraftstoffs = argument.HEIZWERT
    w.titel = argument.TITEL

    resultmsg = w.erzeuge_diagramm(pfad=argument.OUTPUT)
    if resultmsg:
        stop(5, resultmsg)
    print("Das Wirkungsgradkennfeld wurde unter {} gespeichert!".format(argument.OUTPUT))
    sys.exit(0)


if __name__ == "__main__":
    main()
